#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "getf.h"

/* Read an input fragment into the passed in buffer buf up to a maximum length
   of len characters.  '\0' is not appended unless an early termination is 
   encountered *and* there is space for an addition '\0' character to be 
   appended. The terminator character, termchar, is not substituted by a '\0' 
   and is in fact presented in the read fragment if consumed.  Unless there 
   is an obvious parameter error, the number of characters read is always 
   returned. */

long getInputFrag (
	/* @out@ */ char * buf, long len,
	int termchar,
	int (*cbGetc) (void * ctx), void * cbGetcCtx) {
long i;
int c;

	if (!buf || len <= 0 || !cbGetc) return -1;

	for (c = (~termchar) | 1, i=0; i < len; i++) {
		if (c == (int) '\0' || c == termchar || EOF == (c = cbGetc (cbGetcCtx))) {
			buf[i] = '\0';
			break;
		}
		buf[i] = (char) c;
	}
	return i;
}

/* Work around an ANSI requirement that casting a function pointer does not
   change its parameters in any way. */

static int xfgetc (void * ctx) {
	return fgetc ((FILE *) ctx);
}

/* Very much like fgets except that it returns the length of the read string 
   or -1 if there was a problem reading the string and it also terminates 
   early if a '\0' is read. A terminating '\0' is always added to the end. */ 

long fgetstr (char * buff, int n, FILE * fp) {
long len;

	if (!buff) return -1;

	if (1 == n) {
		buff[0] = '\0';
		return 0;
	}

	if (n < 1 || !fp) return -1;

	len = getInputFrag (buff, n-1, (int) '\n', xfgetc, fp);
	if (0 <= len) buff[len] = '\0';
	return len;
}

/* Fills *p with a malloced buffer that contains an input line. The line is 
   always '\0' terminated. If p is NULL, it is not filled in. The number of 
   characters read is returned.  If stdin fails for some reason, partial
   data will be returned and fgetc will set the errno value as expected. If
   insufficient memory is available, *p is not set and 0 is returned. */

size_t fgetstralloc (char ** p, FILE * fp) {
size_t mlen, slen, dlen;
long len;
char * s, * t;

	if (!fp) return (size_t) -1;

	if (!p) {
		char blk[64];
		for (slen = 0; ; slen += 64) {
			len = getInputFrag (blk, 64, (int) '\n', xfgetc, fp);
			if (0 > len) return slen;
			if (64 != len || '\n' == blk[len-1] || '\0' == blk[len-1]) return slen + (size_t) len;
		}
	}

	mlen = 4;
	slen = 0;
	s = NULL;

	for (;;) {
		mlen += mlen;
		if (NULL == (t = realloc (s, mlen))) break;
		len = getInputFrag ((s = t) + slen, dlen = mlen - slen - 1, (int) '\n', xfgetc, fp);
		if (0 > len) len = 0;

		slen += (size_t) len;
		if ((size_t) len < dlen || t[slen-1] == '\n' || t[slen-1] == '\0') {
			s[slen] = '\0';
			*p = s;
			return slen;
		}
	}

	free (s);
	return (size_t) 0;
}

